//-----------------------------------------------------------------------------
// File: wipe.cpp
//
// Desc: DirectX window application created by the DirectX AppWizard
//-----------------------------------------------------------------------------
#define STRICT
#include <windows.h>
#include <basetsd.h>
#include <math.h>
#include <stdio.h>
#include <D3DX8.h>
#include <DXErr8.h>
#include <tchar.h>
#include "D3DApp.h"
#include "D3DFont.h"
#include "D3DUtil.h"
#include "DXUtil.h"
#include "resource.h"
#include "wipe.h"



//-----------------------------------------------------------------------------
// Global access to the app (needed for the global WndProc())
//-----------------------------------------------------------------------------
CMyD3DApplication* g_pApp  = NULL;
HINSTANCE          g_hInst = NULL;




//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
    CMyD3DApplication d3dApp;

    g_pApp  = &d3dApp;
    g_hInst = hInst;

    if( FAILED( d3dApp.Create( hInst ) ) )
        return 0;

    return d3dApp.Run();
}




//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
{
    m_dwCreationWidth           = 500;
    m_dwCreationHeight          = 375;
    m_strWindowTitle            = TEXT( "Stencil Buffer: Scene Wipe" );
    m_bUseDepthBuffer           = TRUE;
    m_dwMinDepthBits    = 15;
    m_dwMinStencilBits  = 1;

    // Create a D3D font using d3dfont.cpp
    m_pFont                     = new CD3DFont( _T("Arial"), 12, D3DFONT_BOLD );
    m_bLoadingApp               = TRUE;
    m_pD3DXMesh                 = NULL;

    ZeroMemory( &m_UserInput, sizeof(m_UserInput) );
    m_fTaurusRotX                = 0.0f;
    m_fTaurusRotX                = 0.0f;
    m_fTeapotRotX                = 0.0f;
    m_fTeapotRotX                = 0.0f;

	m_wipeDelay					= 2.0f;
	m_wipeDir					= 1;
	m_curScene					= 0;

    // Read settings from registry
    ReadSettings();
}




//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
    // TODO: perform one time initialization

    // Drawing loading status message until app finishes loading
    SendMessage( m_hWnd, WM_PAINT, 0, 0 );

    m_bLoadingApp = FALSE;

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: ReadSettings()
// Desc: Read the app settings from the registry
//-----------------------------------------------------------------------------
VOID CMyD3DApplication::ReadSettings()
{
    HKEY hkey;
    if( ERROR_SUCCESS == RegCreateKeyEx( HKEY_CURRENT_USER, DXAPP_KEY, 
        0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hkey, NULL ) )
    {
        // TODO: change as needed

        // Read the stored window width/height.  This is just an example,
        // of how to use DXUtil_Read*() functions.
        DXUtil_ReadIntRegKey( hkey, TEXT("Width"), &m_dwCreationWidth, m_dwCreationWidth );
        DXUtil_ReadIntRegKey( hkey, TEXT("Height"), &m_dwCreationHeight, m_dwCreationHeight );

        RegCloseKey( hkey );
    }
}




//-----------------------------------------------------------------------------
// Name: WriteSettings()
// Desc: Write the app settings to the registry
//-----------------------------------------------------------------------------
VOID CMyD3DApplication::WriteSettings()
{
    HKEY hkey;
    DWORD dwType = REG_DWORD;
    DWORD dwLength = sizeof(DWORD);

    if( ERROR_SUCCESS == RegCreateKeyEx( HKEY_CURRENT_USER, DXAPP_KEY, 
        0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hkey, NULL ) )
    {
        // TODO: change as needed

        // Write the window width/height.  This is just an example,
        // of how to use DXUtil_Write*() functions.
        DXUtil_WriteIntRegKey( hkey, TEXT("Width"), m_rcWindowClient.right );
        DXUtil_WriteIntRegKey( hkey, TEXT("Height"), m_rcWindowClient.bottom );

        RegCloseKey( hkey );
    }
}





//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the display device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{
    BOOL bCapsAcceptable;

    // TODO: Perform checks to see if these display caps are acceptable.
    bCapsAcceptable = TRUE;

    if( bCapsAcceptable )         
        return S_OK;
    else
        return E_FAIL;
}




//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
    // Init the font
    m_pFont->InitDeviceObjects( m_pd3dDevice );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restores scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
    // TODO: setup render states

    // Set up the textures
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_MODULATE );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );

    // Set miscellaneous render states
    m_pd3dDevice->SetRenderState( D3DRS_DITHERENABLE,   FALSE );
    m_pd3dDevice->SetRenderState( D3DRS_SPECULARENABLE, FALSE );
    m_pd3dDevice->SetRenderState( D3DRS_ZENABLE,        TRUE );
    m_pd3dDevice->SetRenderState( D3DRS_AMBIENT,        0x000F0F0F );

    // Set the world matrix
    D3DXMATRIX matIdentity;
    D3DXMatrixIdentity( &matIdentity );
    m_pd3dDevice->SetTransform( D3DTS_WORLD,  &matIdentity );

    // Set up our view matrix. A view matrix can be defined given an eye point,
    // a point to lookat, and a direction for which way is up. Here, we set the
    // eye five units back along the z-axis and up three units, look at the
    // origin, and define "up" to be in the y-direction.
    D3DXMATRIX matView;
    D3DXVECTOR3 vFromPt   = D3DXVECTOR3( 0.0f, 0.0f, -5.0f );
    D3DXVECTOR3 vLookatPt = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
    D3DXVECTOR3 vUpVec    = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );
    D3DXMatrixLookAtLH( &matView, &vFromPt, &vLookatPt, &vUpVec );
    m_pd3dDevice->SetTransform( D3DTS_VIEW, &matView );

    // Set the projection matrix
    D3DXMATRIX matProj;
    FLOAT fAspect = ((FLOAT)m_d3dsdBackBuffer.Width) / m_d3dsdBackBuffer.Height;
    D3DXMatrixPerspectiveFovLH( &matProj, D3DX_PI/4, fAspect, 1.0f, 100.0f );
    m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );

    // Set up lighting states
    D3DLIGHT8 light;
    D3DUtil_InitLight( light, D3DLIGHT_DIRECTIONAL, -1.0f, -1.0f, 2.0f );
    m_pd3dDevice->SetLight( 0, &light );
    m_pd3dDevice->LightEnable( 0, TRUE );
    m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE );

    // Restore the font
    m_pFont->RestoreDeviceObjects();
    
	HRESULT hr;

	// Create a teapot and torus mesh using D3DX
    if( FAILED( hr = D3DXCreateTeapot( m_pd3dDevice, &m_pD3DXMesh, NULL ) ) )
        return DXTRACE_ERR_NOMSGBOX( "D3DXCreateTeapot", hr );
    if( FAILED( hr = D3DXCreateTorus( m_pd3dDevice, 0.7f,1.0f,10,10,&m_pD3DXMesh2, NULL ) ) )
        return DXTRACE_ERR_NOMSGBOX( "D3DXCreateTorus", hr );
	if (FAILED(hr=D3DXCreateTextureFromFile(m_pd3dDevice,"mask.dds",&m_pMask)))
        return DXTRACE_ERR_NOMSGBOX( "LoadMask", hr );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
	// update teapot rotation
	m_fTeapotRotX+=m_fElapsedTime*1.15f;
	m_fTeapotRotY+=m_fElapsedTime*0.4f;

	// update taurus rotation
	m_fTaurusRotX-=m_fElapsedTime*0.5f;
	m_fTaurusRotY-=m_fElapsedTime*0.9f;

	// are we currently in a transition?
	if (m_curScene!=1) {

		// no, decrement countdown
		m_wipeDelay-=m_fElapsedTime;

		// is it time for a new transition?
		if (m_wipeDelay<=0.0f) {

			// yes, initialize wipe timer
			m_wipeTime=0.0f;

			// set scene transition
			m_curScene=1;
		}
	} else {

		// yes, increment wipe position
		m_wipeTime+=m_fElapsedTime;

		// is the wipe finished
		if (m_wipeTime>=1.0f) {

			// yes, set the time until the next wipe
			m_wipeDelay=2.0f;

			// set new scene number
			m_curScene=((m_wipeDir==1)?2:0);

			// set direction of next wipe
			m_wipeDir*=-1;
		}
	}

    return S_OK;
}

typedef struct _rhwTUVert{
	float x;
	float y;
	float z;
	float rhw;
	float u;
	float v;
	_rhwTUVert(float fx,float fy,float fu,float fv){
		x=fx;y=fy;z=0.1f;rhw=0.1f;u=fu;v=fv;
	};
	_rhwTUVert(){};
} rhwTUVert;

#define FVF_rhwTUVert		D3DFVF_XYZRHW|D3DFVF_TEX1

typedef struct _rhwDifVert{
	float x;
	float y;
	float z;
	float rhw;
	DWORD col;
	_rhwDifVert(float fx,float fy,DWORD dwCol){
		x=fx;y=fy;z=0.1f;rhw=0.1f;col=dwCol;
	};
	_rhwDifVert(){};
} rhwDifVert;

#define FVF_rhwDifVert		D3DFVF_XYZRHW|D3DFVF_DIFFUSE

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::Render()
{
    // Clear the viewport
    m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER|D3DCLEAR_STENCIL,
                         0x000000ff, 1.0f, 0L );

    // Begin the scene
    if( SUCCEEDED( m_pd3dDevice->BeginScene() ) )
    {
		// are we wiping?
		if (m_curScene==1) {
			// yes, render the mask in the appropriate direction
			if (m_wipeDir>0)
				RenderMask(m_wipeTime);
			else
				RenderMask(1.0f-m_wipeTime);
		}

		// if the first scene is wholly or partially visible, render it
		// note that stencil comparison for first scene already set in RenderMask()
		if (m_curScene!=2) 
			RenderFirstScene();

		// are we in a transition?
		if (m_curScene==1) 
			
			// yes, change comparison to mask second scene
		    m_pd3dDevice->SetRenderState( D3DRS_STENCILFUNC,D3DCMP_NOTEQUAL);

		// if the second scene is wholly or partially visible, render it
		if (m_curScene!=0) 
			RenderSecondScene();

		// disable stencil function
		m_pd3dDevice->SetRenderState( D3DRS_STENCILENABLE,FALSE);
		m_pd3dDevice->SetRenderState( D3DRS_STENCILFUNC,D3DCMP_ALWAYS);

        // Render stats and help text  
        RenderText();

        // End the scene.
        m_pd3dDevice->EndScene();
    }

    return S_OK;
}


//-----------------------------------------------------------------------------
// Name: RenderText()
// Desc: Renders stats and help text to the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RenderText()
{
    D3DCOLOR fontColor        = D3DCOLOR_ARGB(255,255,255,0);
    D3DCOLOR fontWarningColor = D3DCOLOR_ARGB(255,0,255,255);
    TCHAR szMsg[MAX_PATH] = TEXT("");

    // Output display stats
    FLOAT fNextLine = 40.0f; 

    lstrcpy( szMsg, m_strDeviceStats );
    fNextLine -= 20.0f;
    m_pFont->DrawText( 2, fNextLine, fontColor, szMsg );

    lstrcpy( szMsg, m_strFrameStats );
    fNextLine -= 20.0f;
    m_pFont->DrawText( 2, fNextLine, fontColor, szMsg );

    // Output statistics & help
    fNextLine = (FLOAT) m_d3dsdBackBuffer.Height; 

     lstrcpy( szMsg, TEXT("Press 'F2' to configure display") );
    fNextLine -= 20.0f; m_pFont->DrawText( 2, fNextLine, fontColor, szMsg );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Overrrides the main WndProc, so the sample can do custom message
//       handling (e.g. processing mouse, keyboard, or menu commands).
//-----------------------------------------------------------------------------
LRESULT CMyD3DApplication::MsgProc( HWND hWnd, UINT msg, WPARAM wParam,
                                    LPARAM lParam )
{
    switch( msg )
    {
        case WM_PAINT:
        {
            if( m_bLoadingApp )
            {
                // Draw on the window tell the user that the app is loading
                // TODO: change as needed
                HDC hDC = GetDC( hWnd );
                TCHAR strMsg[MAX_PATH];
                wsprintf( strMsg, TEXT("Loading... Please wait") );
                RECT rct;
                GetClientRect( hWnd, &rct );
                DrawText( hDC, strMsg, -1, &rct, DT_CENTER|DT_VCENTER|DT_SINGLELINE );
                ReleaseDC( hWnd, hDC );
            }
            break;
        }

    }

    return CD3DApplication::MsgProc( hWnd, msg, wParam, lParam );
}




//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Invalidates device objects.  
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
    // TODO: Cleanup any objects created in RestoreDeviceObjects()
    m_pFont->InvalidateDeviceObjects();
    SAFE_RELEASE( m_pD3DXMesh );
    SAFE_RELEASE( m_pD3DXMesh2 );
	SAFE_RELEASE(m_pMask);
    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.  
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
    // TODO: Cleanup any objects created in InitDeviceObjects()
    m_pFont->DeleteDeviceObjects();


    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
    // Cleanup D3D font
    SAFE_DELETE( m_pFont );

    // Write the settings to the registry
    WriteSettings();

    return S_OK;
}

void CMyD3DApplication::RenderMask(float maskTime)
{
	// disable depth buffering
	m_pd3dDevice->SetRenderState( D3DRS_ZENABLE,FALSE);
	m_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE,FALSE);

	// use alpha blending with src=zero and dest=one to prevent writes to color buffer
	m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE,TRUE);
	m_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,D3DBLEND_ZERO);
	m_pd3dDevice->SetRenderState( D3DRS_DESTBLEND,D3DBLEND_ONE);

	// set up alpha testing
	m_pd3dDevice->SetRenderState( D3DRS_ALPHATESTENABLE,TRUE);
	m_pd3dDevice->SetRenderState( D3DRS_ALPHAREF,0x7f);
	m_pd3dDevice->SetRenderState( D3DRS_ALPHAFUNC,D3DCMP_GREATER);

	// set up stencil function to increment stencil of pixels that pass alpha testing
	m_pd3dDevice->SetRenderState( D3DRS_STENCILPASS,D3DSTENCILOP_INCR);
	m_pd3dDevice->SetRenderState( D3DRS_STENCILENABLE,TRUE);
	m_pd3dDevice->SetRenderState( D3DRS_STENCILFUNC,D3DCMP_ALWAYS);

	// disable lighting
	m_pd3dDevice->SetRenderState( D3DRS_LIGHTING,FALSE);

	// set texture wrapping flags
	// prevent wrapping on u axis to avoid blending of opaque and transparent edges
	// allow wrapping on v axis so we can shift the pattern vertically
	m_pd3dDevice->SetTextureStageState(0,D3DTSS_ADDRESSU,D3DTADDRESS_CLAMP);
	m_pd3dDevice->SetTextureStageState(0,D3DTSS_ADDRESSV,D3DTADDRESS_WRAP);

	// select the mask texture
	m_pd3dDevice->SetTexture(0,m_pMask);

	// set vertex shader for rendering the mask primitives
	m_pd3dDevice->SetVertexShader(FVF_rhwTUVert);

	// create wipe pattern primitives and render
	rhwTUVert verts[8];
	float x2,x3;
	float u1=0.0f,u2=1.0f;
	x3=m_d3dsdBackBuffer.Width*maskTime*2.0f;
	x2=x3-m_d3dsdBackBuffer.Width;
	if (x3>m_d3dsdBackBuffer.Width) {
		u2=(m_d3dsdBackBuffer.Width-x2)/(float) m_d3dsdBackBuffer.Width;
		x3=m_d3dsdBackBuffer.Width;
	} else if (x2<0) {
		u1=(m_d3dsdBackBuffer.Width-x3)/(float) m_d3dsdBackBuffer.Width;
		x2=0;
	}
	// note alternating use of maskTime*maskTime used to add a twisting effect to the wipe
	// for a straight horizontal wipe, replace occurances of maskTime*maskTime with maskTime
	verts[0]=rhwTUVert(0.0f,m_d3dsdBackBuffer.Height,u1,maskTime+1.0f);
	verts[1]=rhwTUVert(0.0f,0.0f,u1,maskTime);
	verts[2]=rhwTUVert(x2,m_d3dsdBackBuffer.Height,u1,maskTime*maskTime+1.0f);
	verts[3]=rhwTUVert(x2,0.0f,u1,maskTime*maskTime);
	verts[4]=rhwTUVert(x3,m_d3dsdBackBuffer.Height,u2,maskTime+1.0f);
	verts[5]=rhwTUVert(x3,0.0f,u2,maskTime*maskTime);
	verts[6]=rhwTUVert(m_d3dsdBackBuffer.Width,m_d3dsdBackBuffer.Height,u2,maskTime*maskTime+1.0f);
	verts[7]=rhwTUVert(m_d3dsdBackBuffer.Width,0.0f,u2,maskTime);

	m_pd3dDevice->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP,6,verts,sizeof(rhwTUVert));

	// re-enable z-buffering
	m_pd3dDevice->SetRenderState( D3DRS_ZENABLE,TRUE);
	m_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE,TRUE);

	// set up stencil states for rendering of first scene
	m_pd3dDevice->SetRenderState( D3DRS_STENCILPASS,D3DSTENCILOP_KEEP);
	m_pd3dDevice->SetRenderState( D3DRS_STENCILFUNC,D3DCMP_EQUAL);

	// re-enable lighting
	m_pd3dDevice->SetRenderState( D3DRS_LIGHTING,TRUE);

	// disable alpha blending and testing
	m_pd3dDevice->SetRenderState( D3DRS_ALPHATESTENABLE,FALSE);
	m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE,FALSE);

	// clear texture
	m_pd3dDevice->SetTexture(0,NULL);
}

void CMyD3DApplication::RenderFirstScene()
{
	// set material
	D3DMATERIAL8 mtrl;
	D3DUtil_InitMaterial( mtrl, 1.0f, 0.0f, 0.0f );
	m_pd3dDevice->SetMaterial( &mtrl );

	// set rotation 
	D3DXMATRIX matWorld;
	D3DXMatrixRotationYawPitchRoll(&matWorld,m_fTeapotRotY,m_fTeapotRotX,0.0f);
	m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

	// Render the teapot mesh
	m_pD3DXMesh->DrawSubset(0);
}

void CMyD3DApplication::RenderSecondScene()
{
	// disable z-buffering
	m_pd3dDevice->SetRenderState( D3DRS_ZENABLE,FALSE);
	m_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE,FALSE);

	// render a different background in this scene as a colored quad
	m_pd3dDevice->SetVertexShader(FVF_rhwDifVert);
	rhwDifVert verts[4];
	verts[0]=rhwDifVert(0.0f,m_d3dsdBackBuffer.Height,0xff7f0000);
	verts[1]=rhwDifVert(0.0f,0.0f,0xff7f0000);
	verts[2]=rhwDifVert(m_d3dsdBackBuffer.Width,m_d3dsdBackBuffer.Height,0xff7f0000);
	verts[3]=rhwDifVert(m_d3dsdBackBuffer.Width,0.0f,0xff7f0000);
	m_pd3dDevice->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP,2,verts,sizeof(rhwDifVert));

	// re-enable z-buffering
	m_pd3dDevice->SetRenderState( D3DRS_ZENABLE,TRUE);
	m_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE,TRUE);

	// set material
	D3DMATERIAL8 mtrl;
	D3DUtil_InitMaterial( mtrl, 0.0f, 1.0f, 1.0f );
	m_pd3dDevice->SetMaterial( &mtrl );

	// set rotation
	D3DXMATRIX matWorld;
	D3DXMatrixRotationYawPitchRoll(&matWorld,m_fTaurusRotY,m_fTaurusRotX,0.0f);
	m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

	// render the taurus mesh
	m_pD3DXMesh2->DrawSubset(0);
}
